from abc import ABC, ABCMeta
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Type,
    Union,
    cast,
    no_type_check,
)

from injector import inject, is_decorated_with_inject
from ninja import NinjaAPI
from ninja.constants import NOT_SET
from ninja.operation import Operation
from ninja.security.base import AuthBase
from ninja.types import DictStrAny

from ninja_extra.exceptions import PermissionDenied
from ninja_extra.operation import PathView
from ninja_extra.permissions import BasePermission
from ninja_extra.permissions.base import OperandHolder
from ninja_extra.shortcuts import fail_silently

from .response import ControllerResponse, Detail, Id, Ok
from .route.route_functions import RouteFunction, RouteFunctionContext
from .router import ControllerRouter


class MissingRouterDecoratorException(Exception):
    pass


def compute_api_route_function(
    base_cls: Type["APIController"], controller: Optional[Type["APIController"]] = None
) -> None:
    controller = controller if controller else base_cls
    for cls_route_function in base_cls.get_route_functions():
        cls_route_function.controller = controller
        controller.add_operation_from_route_function(cls_route_function)


class APIControllerModelMetaclass(ABCMeta):
    @no_type_check
    def __new__(mcs, name: str, bases: tuple, namespace: dict):
        cls = super().__new__(mcs, name, bases, namespace)
        if name == "APIController" and ABC in bases:
            return cls

        cls = cast(Type[APIController], cls)
        cls._path_operations = {}
        cls.api = namespace.get("api", None)
        cls.registered = False
        cls.permission_classes = None

        if not namespace.get("tags"):
            tag = str(cls.__name__).lower().replace("controller", "")
            cls.tags = [tag]

        if len(bases) > 1:
            for base_cls in reversed(bases):
                if issubclass(base_cls, APIController):
                    compute_api_route_function(base_cls, cls)

        compute_api_route_function(cls)
        if not is_decorated_with_inject(cls.__init__):
            fail_silently(inject, constructor_or_class=cls)
        return cls


class APIController(ABC, metaclass=APIControllerModelMetaclass):
    # TODO: implement csrf on route function or on controller level. Which can override api csrf
    #   controller should have a csrf ON unless turned off by api instance
    auto_import = (
        True  # set to false and it would be ignored when api.auto_discover is called
    )

    # `_path_operations` a converted dict of APIController route function used by Django-Ninja library
    _path_operations: Dict[str, PathView]
    # `api` a reference to NinjaExtraAPI on APIController registration
    api: Optional[NinjaAPI] = None
    # `auth` primarily defines APIController route function global authentication method.
    auth: Optional[AuthBase] = None
    # `registered` prevents controllers from being register twice or exist in two different `api` instances
    registered: bool
    # `_router` a reference to ControllerRouter of any APIController with a ControllerRouter decorator
    _router: Optional[ControllerRouter] = None
    # `permission_classes` primarily holds permission defined by the ControllerRouter and its used as
    # a fallback if route functions has no permissions definition
    permission_classes: Union[List[Type[BasePermission]], List[OperandHolder[Any]]]
    # `tags` is a property for grouping endpoint in Swagger API docs
    tags: List[str] = []

    # `context` variable will change based on the route function called on the APIController
    # that way we can get some specific items things that belong the route function during execution
    context: RouteFunctionContext = RouteFunctionContext()

    Ok = Ok
    Id = Id
    Detail = Detail

    @classmethod
    def get_router(cls) -> ControllerRouter:
        if not cls._router:
            raise MissingRouterDecoratorException(
                "Controller Router not found. "
                "Did you forget to use the `router` decorator"
            )
        return cls._router

    @classmethod
    def get_path_operations(cls) -> DictStrAny:
        return cls._path_operations

    @classmethod
    def add_operation_from_route_function(cls, route_function: RouteFunction) -> None:
        # converts route functions to Operation model
        cls.add_api_operation(
            view_func=route_function.as_view, **route_function.route.route_params.dict()
        )

    @classmethod
    def add_api_operation(
        cls,
        path: str,
        methods: List[str],
        view_func: Callable,
        *,
        auth: Any = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
    ) -> Operation:
        if path not in cls._path_operations:
            path_view = PathView()
            cls._path_operations[path] = path_view
        else:
            path_view = cls._path_operations[path]
        operation = path_view.add_operation(
            path=path,
            methods=methods,
            view_func=view_func,
            auth=auth or cls.auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
        )
        return operation

    @classmethod
    def get_route_functions(cls) -> Iterator[RouteFunction]:
        for method in cls.__dict__.values():
            if isinstance(method, RouteFunction):
                yield method

    @classmethod
    def permission_denied(cls, permission: BasePermission) -> None:
        message = getattr(permission, "message", None)
        raise PermissionDenied(message)

    def get_permissions(self) -> Iterator[BasePermission]:
        """
        Instantiates and returns the list of permissions that this view requires.
        """
        for permission_class in self.context.permission_classes:
            permission_instance = permission_class()
            yield permission_instance

    def check_permissions(self) -> None:
        """
        Check if the request should be permitted.
        Raises an appropriate exception if the request is not permitted.
        """
        for permission in self.get_permissions():
            if self.context.request and not permission.has_permission(
                request=self.context.request, controller=self
            ):
                self.permission_denied(permission)

    def check_object_permissions(self, obj: Any) -> None:
        """
        Check if the request should be permitted for a given object.
        Raises an appropriate exception if the request is not permitted.
        """
        for permission in self.get_permissions():
            if self.context.request and not permission.has_object_permission(
                request=self.context.request, controller=self, obj=obj
            ):
                self.permission_denied(permission)

    def create_response(
        self, message: Any, status_code: int = 200
    ) -> ControllerResponse:
        return self.Detail(message=message, status_code=status_code)



from typing import Any, List, Optional, Tuple, Type, Union, cast

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from injector import Injector, Module

from ninja_extra.apps import NinjaExtraConfig
from ninja_extra.shortcuts import fail_silently

__all__ = ["service_resolver", "get_injector", "register_injector_modules"]


def get_injector() -> Injector:
    app = cast(
        Optional[NinjaExtraConfig],
        fail_silently(apps.get_app_config, NinjaExtraConfig.name),
    )
    if not app:
        raise ImproperlyConfigured(
            "ninja_extra app is not installed. Did you forget register `ninja_extra` in `INSTALLED_APPS`"
        )
    injector = app.injector
    return injector


def service_resolver(*services: type) -> Union[Tuple[Any], Any]:
    assert services, "Service can not be empty"

    injector = get_injector()

    if len(services) > 1:
        services_resolved: List[Any] = []
        for service in services:
            services_resolved.append(injector.get(service))
        return tuple(services_resolved)
    return injector.get(services[0])


def register_injector_modules(*modules: Union[Module, Type[Module]]) -> None:
    for module in modules:
        injector = get_injector()
        if isinstance(module, type) and issubclass(module, Module):
            module = module()
        injector.binder.install(module)



from django.http import HttpRequest
from django.http.response import HttpResponse
from ninja.operation import PathView as NinjaPathView


class PathView(NinjaPathView):
    async def _async_view(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:  # type: ignore
        return await super(PathView, self)._async_view(request, *args, **kwargs)

    def _sync_view(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:  # type: ignore
        return super(PathView, self)._sync_view(request, *args, **kwargs)



import json
from datetime import datetime, timedelta

import pytest

from ninja_extra.testing import TestClient

from .controllers import EventController, EventSchema
from .models import Event


@pytest.mark.django_db
class TestEventController:
    dummy_data = dict(
        title="TestEvent1Title",
        start_date=str(datetime.now().date()),
        end_date=str((datetime.now() + timedelta(days=5)).date()),
    )

    def test_create_event_works(self):
        client = TestClient(EventController)
        response = client.post(
            "/create",
            json=dict(
                title="TestEvent1Title",
                start_date=str(datetime.now().date()),
                end_date=str((datetime.now() + timedelta(days=5)).date()),
            ),
        )
        assert response.status_code == 201
        data = response.json()
        assert "id" in data
        assert Event.objects.filter(pk=data.get("id")).exists()

    def test_list_events_works(self):
        for i in range(3):
            object_data = self.dummy_data.copy()
            object_data.update(title=f"{object_data['title']}_{i+1}")
            Event.objects.create(**object_data)

        client = TestClient(EventController)
        response = client.get("")
        assert response.status_code == 200
        data = response.json()
        assert len(data) == 3
        event_schema = [
            json.loads(EventSchema.from_orm(item).json())
            for item in Event.objects.all()
        ]
        assert event_schema == data

    def test_get_event_works(self):
        object_data = self.dummy_data.copy()
        object_data.update(title=f"{object_data['title']}_get")

        event = Event.objects.create(**object_data)
        client = TestClient(EventController)
        response = client.get(f"/{event.id}")
        assert response.status_code == 200
        data = response.json()
        event_schema = json.loads(EventSchema.from_orm(event).json())
        assert event_schema == data



from .response import (
    PaginatedResponseSchema,
    RouteParameter,
    get_paginated_response_schema,
)

__all__ = ["PaginatedResponseSchema", "get_paginated_response_schema", "RouteParameter"]



import pytest
from pydantic import ValidationError

from ninja_extra.conf import settings


class CustomPaginationImport:
    pass


class CustomModuleImport:
    pass


def test_setting_imports_string_works(monkeypatch):
    with monkeypatch.context() as m:
        m.setattr(
            settings,
            "INJECTOR_MODULES",
            [
                "tests.test_settings.CustomModuleImport",
            ],
        )
        m.setattr(
            settings, "PAGINATION_CLASS", "tests.test_settings.CustomPaginationImport"
        )

        assert isinstance(settings.INJECTOR_MODULES[0](), CustomModuleImport)
        assert isinstance(settings.PAGINATION_CLASS(), CustomPaginationImport)

    with pytest.raises(ValidationError):
        monkeypatch.setattr(
            settings, "PAGINATION_CLASS", ["tests.test_settings.CustomModuleImport"]
        )
    with pytest.raises(ValidationError):
        monkeypatch.setattr(
            settings, "INJECTOR_MODULES", "tests.test_settings.CustomModuleImport"
        )



"""Django Ninja Extra - Class Based Utility and more for Django Ninja(Fast Django REST framework)"""

__version__ = "0.12.2"

from ninja_extra.controllers import APIController
from ninja_extra.controllers.route import route
from ninja_extra.controllers.router import router
from ninja_extra.dependency_resolver import get_injector, service_resolver
from ninja_extra.main import NinjaExtraAPI

default_app_config = "ninja_extra.apps.NinjaExtraConfig"

__all__ = [
    "NinjaExtraAPI",
    "route",
    "APIController",
    "router",
    "permissions",
    "exceptions",
    "status",
    "shortcuts",
    "get_injector",
    "service_resolver",
    "lazy",
]



from importlib import import_module
from typing import Callable, Optional, Sequence, Type, Union

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest, HttpResponse
from django.utils.module_loading import module_has_submodule
from ninja import NinjaAPI
from ninja.constants import NOT_SET
from ninja.parser import Parser
from ninja.renderers import BaseRenderer

from ninja_extra.controllers.base import APIController
from ninja_extra.controllers.router import ControllerRegistry, ControllerRouter
from ninja_extra.exceptions import APIException

__all__ = [
    "NinjaExtraAPI",
]


class NinjaExtraAPI(NinjaAPI):
    def __init__(
        self,
        *,
        title: str = "NinjaExtraAPI",
        version: str = "1.0.0",
        description: str = "",
        openapi_url: Optional[str] = "/openapi.json",
        docs_url: Optional[str] = "/docs",
        urls_namespace: Optional[str] = None,
        csrf: bool = False,
        auth: Union[Sequence[Callable], Callable, object] = NOT_SET,
        renderer: Optional[BaseRenderer] = None,
        parser: Optional[Parser] = None,
    ) -> None:
        super(NinjaExtraAPI, self).__init__(
            title=title,
            version=version,
            description=description,
            openapi_url=openapi_url,
            docs_url=docs_url,
            urls_namespace=urls_namespace,
            csrf=csrf,
            auth=auth,
            renderer=renderer,
            parser=parser,
        )

        @self.exception_handler(APIException)
        def api_exception_handler(
            request: HttpRequest, exc: APIException
        ) -> HttpResponse:
            message = (
                {"message": exc.message}
                if not isinstance(exc.message, dict)
                else exc.message
            )
            return self.create_response(
                request,
                message,
                status=exc.status_code,
            )

    def register_controllers(self, *controllers: Type[APIController]) -> None:
        for controller in controllers:
            if not issubclass(controller, APIController):
                raise ImproperlyConfigured(
                    f"{controller.__class__.__name__} class is not a controller"
                )
            controller_ninja_router = controller.get_router()
            if not controller.registered and controller_ninja_router:
                self._routers.extend(controller_ninja_router.build_routers())  # type: ignore
                controller_ninja_router.set_api_instance(self)
                controller.registered = True

    def add_controller_router(
        self, *routers: Union[ControllerRouter, Type[ControllerRouter]]
    ) -> None:
        for router in routers:
            if not isinstance(router, ControllerRouter):
                raise ImproperlyConfigured(
                    f"{router.__class__.__name__} class is not a ControllerRouter"
                )
            if not router.controller:
                raise ImproperlyConfigured(
                    f"{router.__class__.__name__} has no controller"
                )
            self.register_controllers(router.controller)

    def auto_discover_controllers(self) -> None:
        from django.apps import apps

        installed_apps = [
            v for k, v in apps.app_configs.items() if not v.name.startswith("django.")
        ]
        possible_module_name = ["api", "controllers"]

        for app_module in installed_apps:
            try:
                app_module_ = import_module(app_module.name)
                for module in possible_module_name:
                    if module_has_submodule(app_module_, module):
                        mod_path = "%s.%s" % (app_module.name, module)
                        import_module(mod_path)
                self.register_controllers(
                    *ControllerRegistry.get_controllers().values()
                )
            except ImportError as ex:
                raise ex



from typing import no_type_check

from django.utils.functional import LazyObject, empty
from django.utils.module_loading import import_string


class LazyStrImport(LazyObject):
    def __init__(self, import_str: str):
        self.__dict__["_import_str"] = import_str
        super(LazyStrImport, self).__init__()

    @no_type_check
    def _setup(self):
        if self._wrapped is empty:
            self._wrapped = import_string(self._import_str)
            return self._wrapped
        return self._wrapped

    @no_type_check
    def __call__(self, *args, **kwargs):
        if self._wrapped is empty:
            self._setup()
        return self._wrapped(*args, **kwargs)



from unittest import mock
from unittest.mock import Mock

import pytest
from django.contrib.auth.models import AnonymousUser, User

from ninja_extra import permissions

anonymous_request = Mock()
anonymous_request.user = AnonymousUser()


class TestPermissionsCompositions:
    @classmethod
    def get_real_user_request(cls):
        _request = Mock()
        user = User.objects.create_user(
            username="eadwin", email="eadwin@example.com", password="password"
        )
        _request.user = user
        return _request

    def test_and_false(self):
        composed_perm = permissions.IsAuthenticated & permissions.AllowAny
        assert composed_perm().has_permission(anonymous_request, None) is False

    @pytest.mark.django_db
    def test_and_true(self):
        request = self.get_real_user_request()
        composed_perm = permissions.IsAuthenticated & permissions.AllowAny
        assert composed_perm().has_permission(request, None) is True

    def test_or_false(self):
        composed_perm = permissions.IsAuthenticated | permissions.AllowAny
        assert composed_perm().has_permission(anonymous_request, None) is True

    @pytest.mark.django_db
    def test_or_true(self):
        request = self.get_real_user_request()
        composed_perm = permissions.IsAuthenticated | permissions.AllowAny
        assert composed_perm().has_permission(request, None) is True

    def test_not_false(self):
        composed_perm = ~permissions.IsAuthenticated
        assert composed_perm().has_permission(anonymous_request, None) is True

    @pytest.mark.django_db
    def test_not_true(self):
        request = self.get_real_user_request()
        composed_perm = ~permissions.AllowAny
        assert composed_perm().has_permission(request, None) is False

    @pytest.mark.django_db
    def test_several_levels_without_negation(self):
        request = self.get_real_user_request()
        composed_perm = (
            permissions.IsAuthenticated
            & permissions.IsAuthenticated
            & permissions.IsAuthenticated
            & permissions.IsAuthenticated
        )
        assert composed_perm().has_permission(request, None) is True

    @pytest.mark.django_db
    def test_several_levels_and_precedence_with_negation(self):
        request = self.get_real_user_request()
        composed_perm = (
            permissions.IsAuthenticated
            & ~permissions.IsAdminUser
            & permissions.IsAuthenticated
            & ~(permissions.IsAdminUser & permissions.IsAdminUser)
        )
        assert composed_perm().has_permission(request, None) is True

    @pytest.mark.django_db
    def test_several_levels_and_precedence(self):
        request = self.get_real_user_request()
        composed_perm = (
            permissions.IsAuthenticated & permissions.IsAuthenticated
            | permissions.IsAuthenticated & permissions.IsAuthenticated
        )
        assert composed_perm().has_permission(request, None) is True

    def test_or_lazyness(self):
        with mock.patch.object(
            permissions.AllowAny, "has_permission", return_value=True
        ) as mock_allow:
            with mock.patch.object(
                permissions.IsAuthenticated, "has_permission", return_value=False
            ) as mock_deny:
                composed_perm = permissions.AllowAny | permissions.IsAuthenticated
                hasperm = composed_perm().has_permission(anonymous_request, None)
                assert hasperm is True
                assert mock_allow.call_count == 1
                mock_deny.assert_not_called()

        with mock.patch.object(
            permissions.AllowAny, "has_permission", return_value=True
        ) as mock_allow:
            with mock.patch.object(
                permissions.IsAuthenticated, "has_permission", return_value=False
            ) as mock_deny:
                composed_perm = permissions.IsAuthenticated | permissions.AllowAny
                hasperm = composed_perm().has_permission(anonymous_request, None)
                assert hasperm is True
                assert mock_deny.call_count == 1
                assert mock_allow.call_count == 1

    def test_object_or_lazyness(self):
        with mock.patch.object(
            permissions.AllowAny, "has_object_permission", return_value=True
        ) as mock_allow:
            with mock.patch.object(
                permissions.IsAuthenticated, "has_object_permission", return_value=False
            ) as mock_deny:
                composed_perm = permissions.AllowAny | permissions.IsAuthenticated
                hasperm = composed_perm().has_object_permission(
                    anonymous_request, None, None
                )
                assert hasperm is True
                assert mock_allow.call_count == 1
                mock_deny.assert_not_called()

        with mock.patch.object(
            permissions.AllowAny, "has_object_permission", return_value=True
        ) as mock_allow:
            with mock.patch.object(
                permissions.IsAuthenticated, "has_object_permission", return_value=False
            ) as mock_deny:
                composed_perm = permissions.IsAuthenticated | permissions.AllowAny
                hasperm = composed_perm().has_object_permission(
                    anonymous_request, None, None
                )
                assert hasperm is True
                assert mock_deny.call_count == 1
                assert mock_allow.call_count == 1

    def test_and_lazyness(self):
        with mock.patch.object(
            permissions.AllowAny, "has_permission", return_value=True
        ) as mock_allow:
            with mock.patch.object(
                permissions.IsAuthenticated, "has_permission", return_value=False
            ) as mock_deny:
                composed_perm = permissions.AllowAny & permissions.IsAuthenticated
                hasperm = composed_perm().has_permission(anonymous_request, None)
                assert hasperm is False
                assert mock_allow.call_count == 1
                assert mock_deny.call_count == 1

        with mock.patch.object(
            permissions.AllowAny, "has_permission", return_value=True
        ) as mock_allow:
            with mock.patch.object(
                permissions.IsAuthenticated, "has_permission", return_value=False
            ) as mock_deny:
                composed_perm = permissions.IsAuthenticated & permissions.AllowAny
                hasperm = composed_perm().has_permission(anonymous_request, None)
                assert hasperm is False
                assert mock_deny.call_count == 1
                mock_allow.assert_not_called()

    def test_object_and_lazyness(self):
        with mock.patch.object(
            permissions.AllowAny, "has_object_permission", return_value=True
        ) as mock_allow:
            with mock.patch.object(
                permissions.IsAuthenticated, "has_object_permission", return_value=False
            ) as mock_deny:
                composed_perm = permissions.AllowAny & permissions.IsAuthenticated
                hasperm = composed_perm().has_object_permission(
                    anonymous_request, None, None
                )
                assert hasperm is False
                assert mock_allow.call_count == 1
                assert mock_deny.call_count == 1

        with mock.patch.object(
            permissions.AllowAny, "has_object_permission", return_value=True
        ) as mock_allow:
            with mock.patch.object(
                permissions.IsAuthenticated, "has_object_permission", return_value=False
            ) as mock_deny:
                composed_perm = permissions.IsAuthenticated & permissions.AllowAny
                hasperm = composed_perm().has_object_permission(
                    anonymous_request, None, None
                )
                assert hasperm is False
                assert mock_deny.call_count == 1
                mock_allow.assert_not_called()



from datetime import date
from typing import List

from django.shortcuts import get_object_or_404
from ninja import Schema

from ninja_extra import APIController, route, router

from .models import Event


class EventSchema(Schema):
    title: str
    start_date: date
    end_date: date

    class Config:
        orm_mode = True


class EventSchemaOut(Schema):
    id: int


@router("events")
class EventController(APIController):
    @route.post(
        "/create", url_name="event-create-url-name", response={201: EventSchemaOut}
    )
    def create_event(self, event: EventSchema):
        event = Event.objects.create(**event.dict())
        return 201, event

    @route.get("", response=List[EventSchema])
    def list_events(self):
        return list(Event.objects.all())

    @route.get("/{int:id}", response=EventSchema)
    def get_event(self, id: int):
        event = get_object_or_404(Event, id=id)
        return event



from unittest import mock

from injector import Binder, Injector, Module

from ninja_extra.dependency_resolver import (
    get_injector,
    register_injector_modules,
    service_resolver,
)


class MyServiceModule(Module):
    def configure(self, binder: Binder) -> None:
        pass


class Service1:
    pass


class Service2:
    pass


def test_get_injector():
    injector = get_injector()
    assert isinstance(injector, Injector)


def test_service_resolver_returns_object_or_list_or_objects():
    service1 = service_resolver(Service1)
    assert isinstance(service1, Service1)

    service1, service2 = service_resolver(Service1, Service2)
    assert isinstance(service1, Service1)
    assert isinstance(service2, Service2)


def test_register_injector_modules_works():
    with mock.patch.object(MyServiceModule, "configure") as mock_configure:
        register_injector_modules(MyServiceModule)
        assert mock_configure.call_count == 1



from .base import APIController, MissingRouterDecoratorException
from .response import Detail, Id, Ok
from .route import Route, RouteInvalidParameterException, route
from .route.route_functions import AsyncRouteFunction, RouteFunction
from .router import ControllerRegistry, ControllerRouter, router

__all__ = [
    "APIController",
    "MissingRouterDecoratorException",
    "route",
    "Route",
    "RouteFunction",
    "AsyncRouteFunction",
    "router",
    "ControllerRegistry",
    "ControllerRouter",
    "RouteInvalidParameterException",
    "Ok",
    "Id",
    "Detail",
]



from django.conf import Settings, settings
from injector import Binder, Module, singleton


class NinjaExtraModule(Module):
    def configure(self, binder: Binder) -> None:
        binder.bind(Settings, to=settings, scope=singleton)  # type: ignore



import sys
from typing import Any, Generic, List, Optional, Type, TypeVar

from ninja import Schema
from ninja.constants import NOT_SET
from pydantic.generics import GenericModel
from pydantic.main import BaseModel
from pydantic.networks import AnyHttpUrl

T = TypeVar("T")

PaginatedResponseSchema = None


class BasePaginatedResponseSchema(Schema):
    count: int
    next: Optional[AnyHttpUrl]
    previous: Optional[AnyHttpUrl]
    results: List[Any]


if sys.version_info >= (3, 8):

    class PaginatedResponseSchema(
        GenericModel, Generic[T], BasePaginatedResponseSchema
    ):
        results: List[T]


def get_paginated_response_schema(
    item_schema: Type[Schema],
) -> Type[BasePaginatedResponseSchema]:
    # fix for paginatedResponseSchema for python 3.6 and 3.7 which doesn't support generic typing
    class ListResponseSchema(BasePaginatedResponseSchema):
        results: List[item_schema]  # type: ignore

    ListResponseSchema.__name__ = f"List{str(item_schema.__name__).capitalize()}"
    return ListResponseSchema


class RouteParameter(BaseModel):
    path: str
    methods: List[str]
    auth: Any = NOT_SET
    response: Any = NOT_SET
    operation_id: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    tags: Optional[List[str]] = None
    deprecated: Optional[bool] = None
    by_alias: bool = False
    exclude_unset: bool = False
    exclude_defaults: bool = False
    exclude_none: bool = False
    url_name: Optional[str] = None
    include_in_schema: bool = True



from unittest.mock import Mock

import django
import pytest
from django.contrib.auth.models import AnonymousUser, User
from ninja import Schema

from ninja_extra import APIController, permissions, route, router
from ninja_extra.controllers import (
    AsyncRouteFunction,
    Detail,
    Id,
    Ok,
    Route,
    RouteFunction,
    RouteInvalidParameterException,
)
from ninja_extra.exceptions import PermissionDenied

anonymous_request = Mock()
anonymous_request.user = AnonymousUser()


@router(
    "permission", permissions=[permissions.IsAuthenticated & permissions.IsAdminUser]
)
class PermissionController(APIController):
    @route.get("/example")
    def example(self):
        return {"message": "OK"}

    @route.get("/example", permissions=[permissions.AllowAny])
    def example_allow_any(self):
        return {"message": "OK"}


class SomeTestController(APIController):
    @route.get("/example")
    def example(self):
        pass

    @route.post("/example")
    def example_post(self):
        pass

    @route.get("/example/{ex_id}")
    def example_get(self, ex_id: str):
        pass

    @route.put("/example/{ex_id}")
    def example_put(self, ex_id: str):
        pass

    @route.delete("/example/{ex_id}")
    def example_delete(self, ex_id: str):
        pass

    @route.generic("/example/list", methods=["POST", "GET"])
    def example_list_create(self, ex_id: str):
        pass


class TestControllerRoutes:
    @pytest.mark.parametrize(
        "path,operation_count",
        [
            ("/example", 2),
            ("/example/{ex_id}", 3),
            ("/example/list", 1),
        ],
    )
    def test_controller_route_build_accurate_operations_list(
        self, path, operation_count
    ):
        path_view = SomeTestController.get_path_operations().get(path)
        assert len(path_view.operations) == operation_count

    def test_controller_route_should_have_an_operation(self):
        for route_func in SomeTestController.get_route_functions():
            path_view = SomeTestController.get_path_operations().get(str(route_func))
            operations = list(
                filter(
                    lambda n: n.operation_id
                    == route_func.route.route_params.operation_id,
                    path_view.operations,
                )
            )
            assert len(operations) == 1
            assert route_func.route.route_params.methods == operations[0].methods

    def test_controller_route_should_right_view_func_type(self):
        assert isinstance(SomeTestController.example, RouteFunction)

    def test_route_generic_invalid_parameters(self):
        with pytest.raises(RouteInvalidParameterException) as ex:

            @route.generic("/example/list", methods=["SOMETHING", "GET"])
            def example_list_create(self, ex_id: str):
                pass

        assert "SOMETHING" in str(ex)

        with pytest.raises(RouteInvalidParameterException) as ex:

            @route.generic("/example/list", methods="SOMETHING")
            def example_list_create(self, ex_id: str):
                pass

        assert "methods must be a list" in str(ex)

    def test_route_response_invalid_parameters(self):
        with pytest.raises(RouteInvalidParameterException) as ex:

            @route.get("/example/list", response=[dict(), ""])
            def example_list_create(self, ex_id: str):
                pass

        assert "Invalid response configuration" in str(ex)

    def test_route_response_parameters_computed_correctly(self):
        unique_response = [Ok, Id, {302: Schema}]
        non_unique_response = [
            Ok,
            Id,
            {201: Schema},
        ]  # Id status_code == 201 so it should be replaced by the dict response

        @route.get("/example/list", response=unique_response)
        def example_unique_response(self, ex_id: str):
            pass

        @route.get("/example/list", response=non_unique_response)
        def example_non_unique_response(self, ex_id: str):
            pass

        assert len(example_unique_response.route.route_params.response) == 3
        assert len(example_non_unique_response.route.route_params.response) == 2

    @pytest.mark.parametrize(
        "func, methods, kwargs",
        [
            (
                "get",
                ["GET"],
                dict(
                    auth="Something",
                    response=dict(),
                    operation_id="operation_id",
                ),
            ),
            (
                "post",
                ["POST"],
                dict(summary="summary", description="description", tags=["dsd"]),
            ),
            (
                "delete",
                ["DELETE"],
                dict(by_alias=True, exclude_unset=True, exclude_defaults=True),
            ),
            ("patch", ["PATCH"], dict(url_name="url_name", include_in_schema=True)),
            ("put", ["PUT"], dict(deprecated=True, exclude_none=True)),
            (
                "generic",
                ["PUT", "PATCH"],
                dict(
                    auth="Something",
                    response=dict(),
                    operation_id="operation_id",
                ),
            ),
        ],
    )
    def test_route_generates_required_route_definitions(self, func, methods, kwargs):
        route_method = getattr(route, func)
        route_instance: Route = (
            route_method("/", methods=methods, **kwargs)
            if func == "generic"
            else route_method("/", **kwargs)
        )
        assert route_instance.route_params.methods == methods
        for k, v in kwargs.items():
            assert getattr(route_instance.route_params, k) == v


@pytest.mark.skipif(django.VERSION < (3, 1), reason="requires django 3.1 or higher")
def test_async_route_function():
    class AsyncSomeTestController(SomeTestController):
        @route.get("/example_async")
        async def example_async(self, ex_id: str):
            pass

    assert isinstance(AsyncSomeTestController.example, RouteFunction)
    assert isinstance(AsyncSomeTestController.example_async, AsyncRouteFunction)


class TestRouteFunction:
    def api_func_with_has_request_param(self, request):
        pass

    def api_func(self):
        pass

    def api_func_with_param(self, example_id: str):
        pass

    async def async_api_func(self):
        pass

    def test_get_required_api_func_signature_return_filtered_signature(self):
        route_function = route.get("")(self.api_func)
        assert not route_function.has_request_param
        sig_inspect, sig_parameter = route_function._get_required_api_func_signature()
        assert len(sig_parameter) == 0

        route_function = route.get("")(self.api_func_with_has_request_param)
        assert route_function.has_request_param
        sig_inspect, sig_parameter = route_function._get_required_api_func_signature()
        assert len(sig_parameter) == 0

        route_function = route.get("")(self.api_func_with_param)
        sig_inspect, sig_parameter = route_function._get_required_api_func_signature()
        assert len(sig_parameter) == 1
        assert str(sig_parameter[0]).replace(" ", "") == "example_id:str"

    def test_from_route_returns_route_function_instance(self):
        route_function = route.get("")(self.api_func)
        assert isinstance(route_function, RouteFunction)

        route_function = route.get("")(self.async_api_func)
        assert isinstance(route_function, AsyncRouteFunction)

    def test_get_controller_init_kwargs(self):
        route_function = route.get("")(self.api_func)
        route_function.controller = Mock()
        route_function.controller.permission_classes = []

        controller_init_kwargs = route_function._get_controller_init_kwargs(
            anonymous_request, "arg1", "arg2", extra="extra"
        )
        assert isinstance(controller_init_kwargs, dict)
        expected_keywords = ("permission_classes", "request", "args", "kwargs")
        for key in expected_keywords:
            assert key in controller_init_kwargs

    def test_get_controller_instance_return_controller_instance(self):
        route_function: RouteFunction = SomeTestController.example
        controller_instance = route_function._get_controller_instance(
            anonymous_request, "arg1", "arg2", extra="extra"
        )
        assert isinstance(controller_instance, SomeTestController)
        assert controller_instance.context.args == ("arg1", "arg2")
        assert controller_instance.context.kwargs == {"extra": "extra"}
        assert controller_instance.context.request == anonymous_request

    def test_process_view_function_result_return_tuple_or_input(self):
        route_function: RouteFunction = SomeTestController.example
        mock_result = Detail("Some Message", status_code=302)
        response = route_function._process_view_function_result(mock_result)
        assert isinstance(response, tuple)
        assert response[1] == mock_result.convert_to_schema()
        assert response[0] == mock_result.status_code

        mock_result = dict(status=302, message="Some Message")
        response = route_function._process_view_function_result(mock_result)
        assert not isinstance(response, tuple)
        assert response == mock_result


@pytest.mark.django_db
class TestAPIControllerRoutePermission:
    @classmethod
    def get_real_user_request(cls):
        _request = Mock()
        user = User.objects.create_user(
            username="eadwin",
            email="eadwin@example.com",
            password="password",
            is_staff=True,
        )
        _request.user = user
        return _request

    def test_route_is_protected_by_global_controller_permission(self):
        with pytest.raises(PermissionDenied) as pex:
            PermissionController.example(anonymous_request)
        assert "You do not have permission to perform this action." in str(
            pex.value.message
        )

    def test_route_protected_by_global_controller_permission_works(self):
        request = self.get_real_user_request()
        response = PermissionController.example(request)
        assert response == {"message": "OK"}

    def test_route_is_protected_by_its_permissions(self):
        response = PermissionController.example_allow_any(anonymous_request)
        assert response == {"message": "OK"}



from typing import Any, cast

from django.apps import AppConfig, apps
from django.utils.translation import gettext_lazy as _
from injector import Injector, Module

from ninja_extra.conf import settings
from ninja_extra.modules import NinjaExtraModule
from ninja_extra.shortcuts import fail_silently


class NinjaExtraConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "ninja_extra"
    verbose_name = _("Django Ninja Extra")
    injector: Injector

    def ready(self) -> None:
        self.injector = Injector([NinjaExtraModule])
        # get django_injector is available or registered
        django_injector_app = fail_silently(
            apps.get_app_config, app_label="django_injector"
        )
        app = cast(Any, django_injector_app)
        if app:
            app.ready()
            self.injector = app.injector
        self.register_injector_modules()

    def register_injector_modules(self) -> None:
        for module in settings.INJECTOR_MODULES:
            if isinstance(module, type) and issubclass(module, Module):
                module = module()
            self.injector.binder.install(module)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ninja_extra',
 'ninja_extra.conf',
 'ninja_extra.controllers',
 'ninja_extra.controllers.route',
 'ninja_extra.permissions',
 'ninja_extra.schemas',
 'ninja_extra.testing']

package_data = \
{'': ['*']}

install_requires = \
['Django >=2.1', 'django-ninja >= 0.15.0', 'injector']

extras_require = \
{'test': ['pytest',
          'pytest-cov',
          'pytest-django',
          'pytest-asyncio',
          'black',
          'isort',
          'injector',
          'flake8',
          'mypy',
          'django-stubs']}

setup(name='django-ninja-extra',
      version='0.12.2',
      description='Django Ninja Extra - Class Based Utility and more for Django Ninja(Fast Django REST framework)',
      author='Ezeudoh Tochukwu',
      author_email='tochukwu.ezeudoh@gmail.com',
      url='https://github.com/eadwinCode/django-ninja-extra',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



from json import dumps as json_dumps
from typing import Any, Callable, Dict, Type
from unittest.mock import Mock
from urllib.parse import urlencode

from ninja.testing.client import NinjaClientBase, NinjaResponse

from ninja_extra import APIController, NinjaExtraAPI


class NinjaExtraClientBase(NinjaClientBase):
    def __init__(self, controller: Type[APIController]) -> None:
        api = NinjaExtraAPI()
        controller_ninja_router = controller.get_router()
        assert controller_ninja_router
        controller_ninja_router.set_api_instance(api)
        self._urls_cache = list(controller_ninja_router.urls_paths(""))
        super(NinjaExtraClientBase, self).__init__(api)

    def request(
        self,
        method: str,
        path: str,
        data: Dict = {},
        json: Any = None,
        **request_params: Any,
    ) -> "NinjaResponse":
        if json is not None:
            request_params["body"] = json_dumps(json)
        if "query" in request_params and isinstance(request_params["query"], dict):
            query = request_params.pop("query")
            url_encode = urlencode(query)
            path = f"{path}?{url_encode}"
        func, request, kwargs = self._resolve(method, path, data, request_params)
        return self._call(func, request, kwargs)  # type: ignore


class TestClient(NinjaExtraClientBase):
    def _call(self, func: Callable, request: Mock, kwargs: Dict) -> "NinjaResponse":
        return NinjaResponse(func(request, **kwargs))


class TestAsyncClient(NinjaExtraClientBase):
    async def _call(self, func: Callable, request: Mock, kwargs: Dict) -> NinjaResponse:
        return NinjaResponse(await func(request, **kwargs))



import abc
from typing import Any, Optional, Type, Union

from ninja import Schema
from pydantic.types import UUID1, UUID3, UUID4, UUID5

from ninja_extra import status


class ControllerResponse(abc.ABC):
    status_code: int = status.HTTP_204_NO_CONTENT

    def __init__(self, **kwargs: Any) -> None:
        pass

    @classmethod
    def get_schema(cls) -> Union[Schema, Type[Schema], Any]:
        raise NotImplementedError

    @abc.abstractmethod
    def convert_to_schema(self) -> Any:
        pass


class Id(ControllerResponse):
    status_code: int = status.HTTP_201_CREATED
    id: Union[int, str, UUID4, UUID1, UUID3, UUID5, Any]

    def __init__(self, id: Any) -> None:
        super(Id, self).__init__()
        self.id = id

    class Id(Schema):
        id: Any

    def convert_to_schema(self) -> Any:
        return self.Id.from_orm(self)

    @classmethod
    def get_schema(cls) -> Union[Schema, Type[Schema], Any]:
        return cls.Id


class Ok(ControllerResponse):
    status_code: int = status.HTTP_200_OK
    message: Any = "Action was successful"

    def __init__(self, message: Optional[Any] = None) -> None:
        super(Ok, self).__init__()
        self.message = message or self.message

    class Ok(Schema):
        message: Any

    def convert_to_schema(self) -> Any:
        return self.Ok.from_orm(self)

    @classmethod
    def get_schema(cls) -> Union[Schema, Type[Schema], Any]:
        return cls.Ok


class Detail(ControllerResponse):
    status_code: int = status.HTTP_200_OK
    message: Any = dict()

    def __init__(
        self, message: Optional[Any] = None, status_code: int = status.HTTP_200_OK
    ) -> None:
        super(Detail, self).__init__()
        self.message = message or self.message
        self.status_code = status_code or self.status_code

    class Detail(Schema):
        message: Any

    def convert_to_schema(self) -> Any:
        return self.Detail.from_orm(self)

    @classmethod
    def get_schema(cls) -> Union[Schema, Type[Schema], Any]:
        return cls.Detail


# class NotFound(ControllerResponse):
#     status_code: int = status.HTTP_404_NOT_FOUND
#     message: Any = 'Item not found'
#
#     def __init__(self, detail: Optional[Any] = None) -> None:
#         super(NotFound, self).__init__()
#         self.detail = detail or self.message
#
#     class NotFound(Schema):
#         message: Any
#
#     def convert_to_schema(self) -> Any:
#         return self.NotFound.from_orm(self)
#
#     @classmethod
#     def get_schema(cls) -> Union[Schema, Type[Schema], Any]:
#         return cls.NotFound



from unittest import mock

import pytest
from django.core.exceptions import ImproperlyConfigured

from ninja_extra import APIController, NinjaExtraAPI, route, router
from ninja_extra.controllers.router import ControllerRegistry


@router("")
class SomeAPIController(APIController):
    @route.get("/example")
    def example(self):
        pass


class InvalidSomeAPIController:
    pass


api = NinjaExtraAPI()
api.register_controllers(SomeAPIController)


@api.get("/global")
def global_op(request):
    pass


def test_api_instance():
    assert len(api._routers) == 2  # default + extra
    for path, rtr in api._routers:
        for path_ops in rtr.path_operations.values():
            for op in path_ops.operations:
                assert op.api is api


def test_api_auto_discover_controller():
    ninja_extra_api = NinjaExtraAPI()
    assert str(SomeAPIController) in ControllerRegistry.get_controllers()

    with mock.patch.object(
        ninja_extra_api, "register_controllers"
    ) as mock_register_controllers:
        ninja_extra_api.auto_discover_controllers()
    assert mock_register_controllers.call_count == 2
    assert (
        "<class 'tests.controllers.EventController'>"
        in ControllerRegistry.get_controllers()
    )

    @router("")
    class SomeAPI2Controller(SomeAPIController):
        auto_import = False

    assert str(SomeAPI2Controller) not in ControllerRegistry.get_controllers()


def test_api_register_controller_works():
    @router("/another")
    class AnotherAPIController(APIController):
        @route.get("/example")
        def example(self):
            pass

    ninja_extra_api = NinjaExtraAPI()
    assert len(ninja_extra_api._routers) == 1
    assert not AnotherAPIController.registered

    ninja_extra_api.register_controllers(AnotherAPIController)
    assert AnotherAPIController.registered
    assert len(ninja_extra_api._routers) == 2

    assert "/another" in {k: v for k, v in ninja_extra_api._routers}

    with pytest.raises(ImproperlyConfigured) as ex:
        ninja_extra_api.register_controllers(InvalidSomeAPIController)

    assert "class is not a controller" in str(ex.value)



import pytest

from ninja_extra import APIController, NinjaExtraAPI, route, router


class RouterControllerExample(APIController):
    auto_import = False

    @route.get("/example")
    def example(self):
        pass


class TestControllerRouter:
    def test_controller_router_as_decorator(self):
        router_instance = router("prefix", tags=["new_tag"])

        @router_instance
        class RouterControllerDecorator(APIController):
            auto_import = False

        assert router_instance.prefix == "prefix"
        assert RouterControllerDecorator.get_router() == router_instance
        assert router_instance.controller == RouterControllerDecorator

        class RouterController(APIController):
            auto_import = False

        router_instance = router("prefix2", tags="new_tag", controller=RouterController)
        assert router_instance.controller == RouterController
        assert router_instance.prefix == "prefix2"
        assert router_instance.tags == ["new_tag"]
        assert RouterController.get_router() == router_instance
        assert router_instance.controller == RouterController

    def test_api_ninja_controller_registration(self):
        router_instance = router("prefix", tags=["new_tag"])
        ninja_extra_api = NinjaExtraAPI()

        @router_instance
        class RouterControllerDecorator(APIController):
            auto_import = False

        ninja_extra_api.add_controller_router(router_instance)
        with pytest.raises(Exception):
            ninja_extra_api.add_controller_router(dict())
        with pytest.raises(Exception):
            ninja_extra_api.add_controller_router(
                router("prefix/test", tags=["new_tag_test"])
            )
